import re


def extract_hashtags(stc):
    tmp = set(part for part in re.split(" ", re.sub("\.|,|;|\(|\)|:|!|^|~|\n", "", stc)) if part.startswith('#'))

    ret_list = []
    for hashtag in tmp:
        i = dict()
        i['text'] = hashtag[1:]
        ret_list.append(i)

    return ret_list


def extract_mentions(stc):
    return set(part for part in re.split("\.|,| |;|\n", re.sub("\.|,|;|!|^|~|\n", "", stc)) if part.startswith('@'))


def get_or_create(session, model, defaults=None, **kwargs):
    instance = session.query(model).filter_by(**kwargs).first()
    if instance:
        return instance
    else:
        params = dict((k, v) for k, v in kwargs.iteritems() if not isinstance(v, ClauseElement))
        if defaults:
            params.update(defaults)
        instance = model(**params)
        return instance


class Dict2Obj(object):
    def __init__(self, dictionary):
        """Constructor"""
        for key in dictionary:
            setattr(self, key, dictionary[key])


from sqlalchemy import Column, Integer, String, DateTime, BigInteger, Boolean
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relationship

Base = declarative_base()


class ESEdge(Base):
    __tablename__ = 'esedges'
    entity = Column(String(256), primary_key=True)
    kind = Column(String(16), primary_key=True)
    status = Column(BigInteger, primary_key=True)

    def __repr__(self):
        return "<entity:{0}, status:{1}, kind:{2}>".format(self.entity,
                                                           self.status,
                                                           self.kind)

    @staticmethod
    def get_esedge(entity_content,
                   status_id,
                   entity_kind):
        return ESEdge(entity=entity_content,
                      status=status_id,
                      kind=entity_kind)

    @staticmethod
    def build_rel(src, kind):
        if src == "Status":
            return relationship("Entity", secondary='esedges',
                                primaryjoin="and_(Status.id==ESEdge.status, ESEdge.kind=='{0}')".format(kind),
                                secondaryjoin="and_(Entity.content==ESEdge.entity, ESEdge.kind==ESEdge.kind)".format(kind))
        elif src == "Entity":
            return relationship("Status",
                                secondary='esedges',
                                primaryjoin="and_(Entity.content==ESEdge.entity, ESEdge.kind=='{0}')".format(kind),
                                secondaryjoin="and_(Status.id==ESEdge.status, ESEdge.kind==ESEdge.kind)")


class EUEdge(Base):
    __tablename__ = 'euedges'

    entity = Column(String(256), primary_key=True)
    kind = Column(String(16), primary_key=True)
    user = Column(BigInteger, primary_key=True)

    def __repr__(self):
        return "<entity:{0}, user:{1}, kind:{2}>".format(self.entity,
                                                         self.user,
                                                         self.kind)

    @staticmethod
    def get_ueedge(entity_content,
                   user_id,
                   entity_kind):
        return EUEdge(entity=entity_content,
                      user=user_id,
                      kind=entity_kind)

    @staticmethod
    def build_rel(src, kind):
        if src == "User":
            return relationship("Entity", secondary='euedges',
                                primaryjoin="and_(User.id==EUEdge.user, EUEdge.kind=='{0}')".format(kind),
                                secondaryjoin="and_(Entity.content==EUEdge.entity, Entity.kind==EUEdge.kind)")
        elif src == "Entity":
            return relationship("User", secondary='euedges',
                                primaryjoin="and_(Entity.content==EUEdge.entity, EUEdge.kind=='{0}')".format(kind),
                                secondaryjoin="and_(User.id==EUEdge.user, EUEdge.kind==EUEdge.kind)".format(kind))


class SSEdge(Base):
    __tablename__ = 'ssedges'

    src_id = Column(BigInteger, primary_key=True)
    dst_id = Column(BigInteger, primary_key=True)
    kind = Column(String(16), primary_key=True)

    def __repr__(self):
        return "<src:{0}, dst:{1}, kind:{2}>".format(self.src_id,
                                                     self.dst_id,
                                                     self.kind)

    @staticmethod
    def get_reply_ssedge(src_id,
                         dst_id):
        return SSEdge(src_id=src_id,
                      dst_id=dst_id,
                      kind="reply")

    @staticmethod
    def get_quote_ssedge(src_id,
                         dst_id):
        return SSEdge(src_id=src_id,
                      dst_id=dst_id,
                      kind="quote")

    @staticmethod
    def j_me(condition, kind):
        tmp = "Status.id==SSEdge.dst_id" if condition == 'primary' else "Status.id==SSEdge.src_id"
        return "and_({0}, SSEdge.kind=='{1}')".format(tmp, kind)

    @staticmethod
    def j_by_me(condition, kind):
        tmp = "Status.id==SSEdge.dst_id" if condition == 'secondary' else "Status.id==SSEdge.src_id"
        return "and_({0}, SSEdge.kind=='{1}')".format(tmp, kind)

    @staticmethod
    def build_rel(condition, kind):
        if condition == "me":
            return relationship("Status", secondary='ssedges', primaryjoin=SSEdge.j_me('primary', kind),
                                secondaryjoin=SSEdge.j_me('secondary', kind))
        elif condition == "by_me":
            return relationship("Status", secondary='ssedges', primaryjoin=SSEdge.j_by_me('primary', kind),
                                secondaryjoin=SSEdge.j_by_me('secondary', kind))


class USEdge(Base):
    __tablename__ = 'usedges'

    user = Column(BigInteger, primary_key=True)
    status = Column(BigInteger, primary_key=True)
    kind = Column(String(16), primary_key=True)

    def __repr__(self):
        return "<user:{0}, status:{1}, kind:{2}>".format(self.user,
                                                         self.status,
                                                         self.kind)

    @staticmethod
    def get_tweet_usedge(user,
                         status):
        return USEdge(user=user,
                      status=status,
                      kind="tweet")

    @staticmethod
    def get_retweet_usedge(user,
                           status):
        return USEdge(user=user,
                      status=status,
                      kind="retweet")

    @staticmethod
    def get_quote_usedge(user,
                         status):
        return USEdge(user=user,
                      status=status,
                      kind="quote")

    @staticmethod
    def get_reply_usedge(user,
                         status):
        return USEdge(user=user,
                      status=status,
                      kind="reply")

    @staticmethod
    def get_favorite_usedge(user,
                            status):
        return USEdge(user=user,
                      status=status,
                      kind="favorite")

    @staticmethod
    def get_mention_usedge(user,
                           status):
        return USEdge(user=user,
                      status=status,
                      kind="mention")

    @staticmethod
    def build_rel(src, kind):
        if src == "User":
            return relationship("Status", secondary='usedges',
                                primaryjoin="and_(User.id==USEdge.user, USEdge.kind=='{0}')".format(kind),
                                secondaryjoin="and_(Status.id==USEdge.status, USEdge.kind=='{0}')".format(kind))
        elif src == "Status":
            return relationship("User", secondary='usedges',
                                primaryjoin="and_(Status.id==USEdge.status, USEdge.kind=='{0}')".format(kind),
                                secondaryjoin="and_(User.id==USEdge.user, USEdge.kind=='{0}')".format(kind))


class UUEdge(Base):
    __tablename__ = 'uuedges'

    src_id = Column(BigInteger, primary_key=True)
    dst_id = Column(BigInteger, primary_key=True)
    kind = Column(String(16), primary_key=True)

    def __repr__(self):
        return "<src:{0}, dst:{1}, kind:{2}>".format(self.src_id,
                                                     self.dst_id,
                                                     self.kind)

    @staticmethod
    def get_follow_uuedge(src_user,
                          dst_user):
        return UUEdge(src_id=src_user,
                      dst_id=dst_user,
                      kind="follow")

    @staticmethod
    def get_favorite_uuedge(src_user,
                            dst_user):
        return UUEdge(src_id=src_user,
                      dst_id=dst_user,
                      kind="favorite")

    @staticmethod
    def get_reply_uuedge(src_user,
                         dst_user):
        return UUEdge(src_id=src_user,
                      dst_id=dst_user,
                      kind="reply")

    @staticmethod
    def get_retweet_uuedge(src_user,
                           dst_user):
        return UUEdge(src_id=src_user,
                      dst_id=dst_user,
                      kind="retweet")

    @staticmethod
    def get_quote_uuedge(src_user,
                         dst_user):
        return UUEdge(src_id=src_user,
                      dst_id=dst_user,
                      kind="quote")

    @staticmethod
    def get_mention_uuedge(src_user,
                           dst_user):
        return UUEdge(src_id=src_user,
                      dst_id=dst_user,
                      kind="mention")

    @staticmethod
    def j_me(condition, kind):
        tmp = "User.id==UUEdge.dst_id" if condition == 'primary' else "User.id==UUEdge.src_id"
        return "and_({0}, UUEdge.kind=='{1}')".format(tmp, kind)

    @staticmethod
    def j_by_me(condition, kind):
        tmp = "User.id==UUEdge.dst_id" if condition == 'secondary' else "User.id==UUEdge.src_id"
        return "and_({0}, UUEdge.kind=='{1}')".format(tmp, kind)

    @staticmethod
    def build_rel(condition, kind):
        if condition == "me":
            return relationship("User", secondary='uuedges',
                                primaryjoin=UUEdge.j_me("primary", kind),
                                secondaryjoin=UUEdge.j_me("secondary", kind))
        elif condition == "by_me":
            return relationship("User", secondary='uuedges',
                                primaryjoin=UUEdge.j_by_me("primary", kind),
                                secondaryjoin=UUEdge.j_by_me("secondary", kind))


class Entity(Base):
    __tablename__ = 'entities'

    content = Column(String(256), primary_key=True)
    kind = Column(String(16), primary_key=True)

    @staticmethod
    def get_new_url(url):
        return Entity(content=url, kind="url")

    @staticmethod
    def get_new_domain(domain):
        return Entity(content=domain, kind="domain")

    @staticmethod
    def get_new_hashtag(hashtag):
        return Entity(content=hashtag, kind="hashtag")

    @staticmethod
    def get_new_media(media):
        return Entity(content=media, kind="media")

    @staticmethod
    def get_new_regex(regex_match):  # this is just in case some day I find it useful
        return Entity(content=regex_match, kind="regex")

    def __repr__(self):
        return "<kind:{0}, content:{1}>".format(self.kind, self.content)

    # >> EUEdge
    u_url_me = EUEdge.build_rel("Entity", "url")
    u_domain_me = EUEdge.build_rel("Entity", "domain")
    u_hashtag_me = EUEdge.build_rel("Entity", "hashtag")
    u_media_me = EUEdge.build_rel("Entity", "media")
    u_regex_me = EUEdge.build_rel("Entity", "regex")

    # >> ESEdge
    s_url_me = ESEdge.build_rel("Entity", "url")
    s_domain_me = ESEdge.build_rel("Entity", "domain")
    s_hashtag_me = ESEdge.build_rel("Entity", "hashtag")
    s_media_me = ESEdge.build_rel("Entity", "media")
    s_regex_me = ESEdge.build_rel("Entity", "regex")


class User(Base):
    __tablename__ = 'users'

    id = Column(BigInteger, primary_key=True)  # Integer   |
    statuses_count = Column(Integer)
    followers_count = Column(Integer)
    followees_count = Column(Integer)  # originally "friends_count"
    favorites_count = Column(Integer)  # originally "favourites_count"
    listed_count = Column(Integer)
    screen_name = Column(String(20))  # String
    name = Column(String(20))
    description = Column(String(256))
    lang = Column(String(8))
    time_zone = Column(String(32))
    location = Column(String(256))
    profile_image_url = Column(String(256))
    default_profile = Column(Boolean)  # Boolean
    default_profile_image = Column(Boolean)
    geo_enabled = Column(Boolean)
    verified = Column(Boolean)
    created_at = Column(DateTime)  # Date Time

    def __repr__(self):
        return "<id:{0}, screen_name:{1}, name:{2}, flwr:{3}, flwe:{4}, lang:{5} desc:{6}, member_since:{7}>".format(
            self.id, self.screen_name, self.name, self.followers_count, self.followees_count, self.lang,
            self.description.replace('\n', ' '), self.created_at)

    # >> UUEdge
    u_followed_me = UUEdge.build_rel("me", "follow")
    u_favorited_me = UUEdge.build_rel("me", "favorite")
    u_replied_me = UUEdge.build_rel("me", "reply")
    u_retweeted_me = UUEdge.build_rel("me", "retweet")
    u_mentioned_me = UUEdge.build_rel("me", "mention")
    u_quoted_me = UUEdge.build_rel("me", "quote")
    u_followed_by_me = UUEdge.build_rel("by_me", "follow")
    u_favorited_by_me = UUEdge.build_rel("by_me", "favorite")
    u_replied_by_me = UUEdge.build_rel("by_me", "reply")
    u_retweeted_by_me = UUEdge.build_rel("by_me", "retweet")
    u_mentioned_by_me = UUEdge.build_rel("by_me", "mention")
    u_quoted_by_me = UUEdge.build_rel("by_me", "quote")

    # >> USEdge
    s_tweeted_by_me = USEdge.build_rel("User", "tweet")
    s_retweeted_by_me = USEdge.build_rel("User", "retweet")
    s_replied_by_me = USEdge.build_rel("User", "reply")
    s_favorited_by_me = USEdge.build_rel("User", "favorite")
    s_quoted_by_me = USEdge.build_rel("User", "quote")
    s_mentioned_me = USEdge.build_rel("User", "mention")

    # >> EUEdge
    e_url_by_me = EUEdge.build_rel("User", "url")
    e_domain_by_me = EUEdge.build_rel("User", "domain")
    e_hashtag_by_me = EUEdge.build_rel("User", "hashtag")
    e_media_by_me = EUEdge.build_rel("User", "media")
    e_regex_by_me = EUEdge.build_rel("User", "regex")


class Status(Base):
    __tablename__ = 'statuses'

    id = Column(BigInteger, primary_key=True)  # Integer
    favorite_count = Column(Integer)  # originally "favourite_count"
    retweet_count = Column(Integer)
    text = Column(String(200))  # String
    lang = Column(String(5))
    coordinates = Column(String(128))
    source = Column(String(128))
    created_at = Column(DateTime)  # DateTime

    def __repr__(self):
        return "<id:{0}, lang:{1}, created_at:{2}, date:{3}>".format(self.id, self.lang, self.text.replace('\n', ' '),
                                                                     self.created_at)

    # >> USEdge
    u_tweeted_me = USEdge.build_rel("Status", "tweet")
    u_retweeted_me = USEdge.build_rel("Status", "retweet")
    u_replied_me = USEdge.build_rel("Status", "reply")
    u_favorited_me = USEdge.build_rel("Status", "favorite")
    u_mentioned_by_me = USEdge.build_rel("Status", "mention")

    # >> SSEdge
    s_replied_me = SSEdge.build_rel("me", 'reply')
    s_replied_by_me = SSEdge.build_rel("by_me", 'reply')
    s_quoted_me = SSEdge.build_rel("me", 'quote')
    s_quoted_by_me = SSEdge.build_rel("by_me", 'quote')

    # >> ESEdge
    e_url_by_me = ESEdge.build_rel("Status", "url")
    e_domain_by_me = ESEdge.build_rel("Status", "domain")
    e_hashtag_by_me = ESEdge.build_rel("Status", "hashtag")
    e_media_by_me = ESEdge.build_rel("Status", "media")
    e_regex = ESEdge.build_rel("Status", "regex")



from SkitterSchema.twitter_schema import User, Entity, Status, UUEdge, USEdge, SSEdge, ESEdge, EUEdge
from SkitterSchema.util import extract_hashtags, Dict2Obj
from urllib.parse import urlparse


class ProcessedOutcome:
    def __init__(self):
        self.schema_users = []
        self.schema_tweets = []
        self.schema_entities = []
        self.schema_edges = []

    def __repr__(self):
        return "{0}\n{1}\n{2}\n{3}\n".format(self.schema_users,
                                             self.schema_tweets,
                                             self.schema_entities,
                                             self.schema_edges)

    def __add__(self, other):

        if type(other) == User:
            self.schema_users += [other]

        if type(other) == Status:
            self.schema_tweets += [other]

        if type(other) == Entity:
            self.schema_entities += [other]

        if type(other) in [UUEdge, USEdge, SSEdge, ESEdge, EUEdge]:
            self.schema_edges += [other]

        if type(other) == ProcessedOutcome:
            self.schema_users += other.schema_users
            self.schema_tweets += other.schema_tweets
            self.schema_entities += other.schema_entities
            self.schema_edges += other.schema_edges

        return self

    def last_user(self):
        return self.schema_users[-1]

    def add_user(self, user):
        self.schema_users.append(user)

    def last_status(self):
        return self.schema_tweets[-1]

    def add_status(self, tweet):
        self.schema_tweets.append(tweet)

    def last_entity(self):
        return self.schema_entities[-1]

    def add_entity(self, entity):
        self.schema_entities.append(entity)

    def last_edge(self):
        return self.schema_edges[-1]

    def add_edge(self, edge):
        self.schema_edges.append(edge)

    def add_to_session(self, session):
        for i in self.schema_users + self.schema_tweets + self.schema_edges + self.schema_entities:
            session.merge(i)

    def get_filtered_edges(self, type_edge, kind, func=lambda x: x):
        edges = []
        for edge in self.schema_edges:
            if type(edge) == type_edge and edge.kind == kind:
                edges.append(func(edge))
        return edges


def compare_user(schema_user, tweepy_user):
    """ Compares a schema User, defined in twitter_schema.py with a tweepy User.
    :param schema_user: TwitterSchema User.
    :param tweepy_user: tweepy User.
    :return: whether all the attributes are the same.
    """

    return \
        schema_user.id == tweepy_user.id and schema_user.statuses_count == tweepy_user.statuses_count and \
        schema_user.followers_count == tweepy_user.followers_count and \
        schema_user.followees_count == tweepy_user.friends_count and \
        schema_user.favorites_count == tweepy_user.favourites_count and \
        schema_user.listed_count == tweepy_user.listed_count and \
        schema_user.screen_name == tweepy_user.screen_name and \
        schema_user.name == tweepy_user.name and \
        schema_user.description == tweepy_user.description and \
        schema_user.lang == tweepy_user.lang and \
        schema_user.time_zone == tweepy_user.time_zone and \
        schema_user.location == tweepy_user.location and \
        schema_user.profile_image_url == tweepy_user.profile_image_url and \
        schema_user.default_profile == tweepy_user.default_profile and \
        schema_user.default_profile_image == tweepy_user.default_profile_image and \
        schema_user.geo_enabled == tweepy_user.geo_enabled and \
        schema_user.verified == tweepy_user.verified


def process_url(url):
    outcome = ProcessedOutcome()
    schema_entity_url = Entity.get_new_url(url="".join(urlparse(url['expanded_url'])[1:]))
    outcome.add_entity(schema_entity_url)
    return outcome


def process_domain(url):
    outcome = ProcessedOutcome()

    schema_entity_domain = Entity.get_new_domain(domain=urlparse(url['expanded_url']).netloc)
    outcome.add_entity(schema_entity_domain)
    return outcome


def process_hashtag(hashtag):
    outcome = ProcessedOutcome()
    schema_entity_hashtag = Entity.get_new_hashtag(hashtag="#" + hashtag['text'])
    outcome.add_entity(schema_entity_hashtag)
    return outcome


def process_media(media):
    outcome = ProcessedOutcome()
    schema_entity_media = Entity.get_new_media(media="".join(urlparse(media['media_url'])[1:]))
    outcome.add_entity(schema_entity_media)
    return outcome


def process_user(user):
    """ This function receives a twitter user and creates a User object that can be stored, as well as its mentions and
    hashtags. Currently, it does not store any of the mentions in the description, as this would require another call
    from the api, but it is something to think about.
    :param user: tweepy User object.
    :return: ProcessedOutcome user object.
    """

    if type(user) is dict:
        user = Dict2Obj(user)

    outcome = ProcessedOutcome()

    # Gets User
    this_user = User(id=user.id, statuses_count=user.statuses_count, followers_count=user.followers_count,
                     followees_count=user.friends_count, favorites_count=user.favourites_count,
                     listed_count=user.listed_count, screen_name=user.screen_name, name=user.name,
                     description=user.description, location=user.location, profile_image_url=user.profile_image_url,
                     time_zone=user.time_zone, lang=user.lang, default_profile=user.default_profile,
                     default_profile_image=user.default_profile_image, geo_enabled=user.geo_enabled,
                     verified=user.verified, created_at=user.created_at)

    outcome.add_user(this_user)

    if 'url' in user.entities and 'urls' in user.entities['url']:
        for url in user.entities['url']['urls']:
            if 'expanded_url' in url and url['expanded_url'] is not None:
                if urlparse(url['expanded_url']).netloc == "twitter.com":
                    continue
                outcome += process_url(url)
                outcome += EUEdge.get_ueedge(outcome.last_entity().content, this_user.id, outcome.last_entity().kind)

                outcome += process_domain(url)
                outcome += EUEdge.get_ueedge(outcome.last_entity().content, this_user.id, outcome.last_entity().kind)

    # Gets Hashtags
    hashtags = extract_hashtags(user.description)
    for hashtag in hashtags:
        outcome += process_hashtag(hashtag)
        outcome += EUEdge.get_ueedge(outcome.last_entity().content, this_user.id, outcome.last_entity().kind)

    return outcome


def process_status(status):
    if type(status) is dict:
        status = Dict2Obj(status)

    outcome = ProcessedOutcome()

    # Initialization - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    # Adds: - tweeting_user
    tmp = process_user(status.user)
    this_user = tmp.schema_users[0]
    outcome += tmp

    # Adds : - tweeted_status (PARTIAL)
    this_status = Status(id=status.id, favorite_count=status.favorite_count, retweet_count=status.retweet_count,
                         text=status.text, lang=status.lang, coordinates=str(status.coordinates),
                         source=status.source, created_at=status.created_at)

    # Gets Retweet (AND TERMINATES) - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    # Adds: - retweeted_status/user/entities/edges
    #       - tweeting_user --rt--> retweeted_status
    #       - tweeting_user --rt--> retweeted_user
    if hasattr(status, 'retweeted_status'):
        # Add status
        outcome += process_status(status.retweeted_status)

        # Add edges
        outcome.add_edge(UUEdge.get_retweet_uuedge(this_user.id, outcome.last_user().id))  # User to user
        outcome.add_edge(USEdge.get_retweet_usedge(this_user.id, outcome.last_status().id))  # User to status

        return outcome

    # Gets Quoted Status - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    # Adds: - quoted_status/user/entities/edges
    #       - tweeting_user  --qt--> quoted_status
    #       - tweeting_user  --qt--> quoted_user
    #       - tweeted_status --qt--> quoted_status
    if hasattr(status, 'quoted_status'):
        outcome += process_status(status.quoted_status)
        outcome += USEdge.get_quote_usedge(this_user.id, outcome.last_status().id)
        outcome += UUEdge.get_quote_uuedge(this_user.id, outcome.last_user().id)
        outcome += SSEdge.get_quote_ssedge(this_status.id, outcome.last_status().id)

    # Gets Normal Tweets, Quotes and Replies - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    # Adds: - tweeted_status (COMPLETES)
    outcome += this_status
    outcome += USEdge.get_tweet_usedge(this_user.id, this_status.id)

    # Adds: - url_entity*/domain_entity*
    #       - this_user   <--> url_entity/domain_entity
    #       - this_status <--> url_entity/domain_entity
    if 'urls' in status.entities:
        for url in status.entities['urls']:
            if 'expanded_url' in url and url['expanded_url'] is not None:
                if urlparse(url['expanded_url']).netloc == "twitter.com":
                    continue
                outcome += process_url(url)
                outcome += EUEdge.get_ueedge(outcome.last_entity().content, this_user.id, outcome.last_entity().kind)
                outcome += ESEdge.get_esedge(outcome.last_entity().content, this_status.id, outcome.last_entity().kind)
                outcome += process_domain(url)
                outcome += EUEdge.get_ueedge(outcome.last_entity().content, this_user.id, outcome.last_entity().kind)
                outcome += ESEdge.get_esedge(outcome.last_entity().content, this_status.id, outcome.last_entity().kind)

    # Adds: - hashtag_entity*
    #       - this_user   <--> hashtag_entity
    #       - this_status <--> hashtag_entity
    if 'hashtags' in status.entities:
        for hashtag in status.entities['hashtags']:
            outcome += process_hashtag(hashtag)
            outcome += EUEdge.get_ueedge(outcome.last_entity().content, this_user.id, outcome.last_entity().kind)
            outcome += ESEdge.get_esedge(outcome.last_entity().content, this_status.id, outcome.last_entity().kind)

    # Adds: - media_entity*
    #       - this_user   <--> media_entity
    #       - this_status <--> media_entity
    if 'media' in status.entities:
        for media in status.entities['media']:
            outcome += process_media(media)
            outcome += EUEdge.get_ueedge(outcome.last_entity().content, this_user.id, outcome.last_entity().kind)
            outcome += ESEdge.get_esedge(outcome.last_entity().content, this_status.id, outcome.last_entity().kind)

    # Adds: - this_status -mt-> mentioned_user
    #       - this_user   -mt-> mentioned_user
    if 'user_mentions' in status.entities:
        for user_mention in status.entities['user_mentions']:
            outcome += UUEdge.get_mention_uuedge(this_user.id, user_mention['id'])
            outcome += USEdge.get_mention_usedge(user_mention['id'], this_status.id)

    # Adds: - this_user -mt-> replied_user
    #       - this_status   -mt-> replied_status
    if hasattr(status, 'in_reply_to_user_id') and status.in_reply_to_user_id is not None:
        outcome += UUEdge.get_reply_uuedge(this_user.id, status.in_reply_to_user_id)
        outcome += SSEdge.get_reply_ssedge(this_status.id, status.in_reply_to_status_id)

    return outcome


def process_timeline(timeline):
    outcome = ProcessedOutcome()

    for status in timeline:
        outcome += process_status(status)

    return outcome


def process_favourites(favorites, user):
    outcome = ProcessedOutcome()

    for status in favorites:
        outcome += process_status(status)
        outcome += UUEdge.get_favorite_uuedge(user.id, outcome.last_user().id)
        outcome += USEdge.get_favorite_usedge(user.id, outcome.last_status().id)

    return outcome


def process_followers(followers, user):
    outcome = ProcessedOutcome()

    for follower in followers:
        outcome += UUEdge.get_follow_uuedge(follower, user.id)

    return outcome


def process_followees(followees, user):
    outcome = ProcessedOutcome()

    for followee in followees:
        outcome += UUEdge.get_follow_uuedge(user.id, followee)

    return outcome



from setuptools import setup

setup(name='SkitterSchema',
      version='0.0.3',
      description='OSN schema for online social networks.',
      url='https://github.com/manoelhortaribeiro/SkitterSchema',
      download_url='https://github.com/manoelhortaribeiro/annotated_bibliography/dist/annotated_bibliography-0.1.tar.gz',
      author='Manoel Horta Ribeiro',
      author_email='manoelhortaribeiro@gmail.com',
      license='MIT',
      packages=['SkitterSchema'],
      keywords='ORM schema online social network',
      install_requires=['sqlalchemy'],
      include_package_data=True,
      zip_safe=False)
